@@ -25,6 +25,7 @@
 #include <osmocom/core/stats.h>
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/stat_item.h>
+#include <osmocom/core/jhash.h>
 
 #include <osmocom/vty/vty.h>
 
@@ -43,6 +44,7 @@
 #include <osmocom/hnbgw/hnbgw_cn.h>
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbgw/mgw_fsm.h>
+#include <osmocom/hnbgw/tdefs.h>
 
 struct hnbgw *g_hnbgw = NULL;
 
@@ -92,92 +94,11 @@ static void hnbgw_store_hnb_rab_durations(void *data)
  * UE Context
  ***********************************************************************/
 
-struct ue_context *ue_context_by_id(uint32_t id)
+uint32_t get_next_ue_ctx_id(void)
 {
-	struct ue_context *ue;
-
-	llist_for_each_entry(ue, &g_hnbgw->ue_list, list) {
-		if (ue->context_id == id)
-			return ue;
-	}
-	return NULL;
-
+	return g_hnbgw->next_ue_ctx_id++;
 }
 
-struct ue_context *ue_context_by_imsi(const char *imsi)
-{
-	struct ue_context *ue;
-
-	llist_for_each_entry(ue, &g_hnbgw->ue_list, list) {
-		if (!strcmp(ue->imsi, imsi))
-			return ue;
-	}
-	return NULL;
-}
-
-struct ue_context *ue_context_by_tmsi(uint32_t tmsi)
-{
-	struct ue_context *ue;
-
-	llist_for_each_entry(ue, &g_hnbgw->ue_list, list) {
-		if (ue->tmsi == tmsi)
-			return ue;
-	}
-	return NULL;
-}
-
-static void ue_context_free_by_hnb(const struct hnb_context *hnb)
-{
-	struct ue_context *ue, *tmp;
-
-	llist_for_each_entry_safe(ue, tmp, &g_hnbgw->ue_list, list) {
-		if (ue->hnb == hnb)
-			ue_context_free(ue);
-	}
-}
-
-static uint32_t get_next_ue_ctx_id(void)
-{
-	uint32_t id;
-
-	do {
-		id = g_hnbgw->next_ue_ctx_id++;
-	} while (ue_context_by_id(id));
-
-	return id;
-}
-
-struct ue_context *ue_context_alloc(struct hnb_context *hnb, const char *imsi,
-				    uint32_t tmsi)
-{
-	struct ue_context *ue;
-
-	ue = talloc_zero(g_hnbgw, struct ue_context);
-	if (!ue)
-		return NULL;
-
-	ue->hnb = hnb;
-	if (imsi)
-		OSMO_STRLCPY_ARRAY(ue->imsi, imsi);
-	else
-		ue->imsi[0] = '\0';
-	ue->tmsi = tmsi;
-	ue->context_id = get_next_ue_ctx_id();
-	llist_add_tail(&ue->list, &g_hnbgw->ue_list);
-
-	LOGP(DHNBAP, LOGL_INFO, "created UE context: id 0x%x, imsi %s, tmsi 0x%x\n",
-	     ue->context_id, imsi? imsi : "-", tmsi);
-
-	return ue;
-}
-
-void ue_context_free(struct ue_context *ue)
-{
-	llist_del(&ue->list);
-	talloc_free(ue);
-}
-
-
 /***********************************************************************
  * HNB Context
  ***********************************************************************/
@@ -231,31 +152,66 @@ static struct hnb_context *hnb_context_alloc(struct osmo_stream_srv_link *link,
 	return ctx;
 }
 
-const char *umts_cell_id_name(const struct umts_cell_id *ucid)
+int umts_cell_id_to_str_buf(char *buf, size_t buflen, const struct umts_cell_id *ucid)
 {
-	const char *fmtstr = "%03u-%02u-L%u-R%u-S%u-C%u";
+	struct osmo_strbuf sb = { .buf = buf, .len = buflen };
+	OSMO_STRBUF_APPEND_NOLEN(sb, osmo_plmn_name_buf, &ucid->plmn);
+	OSMO_STRBUF_PRINTF(sb, "-L%u-R%u-S%u-C%u", ucid->lac, ucid->rac, ucid->sac, ucid->cid);
+	return sb.chars_needed;
+}
 
-	if (g_hnbgw->config.plmn.mnc_3_digits)
-		fmtstr = "%03u-%03u-L%u-R%u-S%u-C%u";
+char *umts_cell_id_to_str_c(void *ctx, const struct umts_cell_id *ucid)
+{
+	OSMO_NAME_C_IMPL(ctx, 64, "ERROR", umts_cell_id_to_str_buf, ucid)
+}
 
-	return talloc_asprintf(OTC_SELECT, fmtstr, ucid->mcc, ucid->mnc, ucid->lac, ucid->rac,
-			       ucid->sac, ucid->cid);
+const char *umts_cell_id_to_str(const struct umts_cell_id *ucid)
+{
+	return umts_cell_id_to_str_c(OTC_SELECT, ucid);
+}
+
+/* Useful to index a hash table by struct umts_cell_id. */
+uint32_t umts_cell_id_hash(const struct umts_cell_id *ucid)
+{
+	return osmo_jhash(ucid, sizeof(*ucid), 0x423423);
 }
 
 /* parse a string representation of an umts_cell_id into its decoded representation */
 int umts_cell_id_from_str(struct umts_cell_id *ucid, const char *instr)
 {
-	int rc = sscanf(instr, "%hu-%hu-L%hu-R%hu-S%hu-C%u", &ucid->mcc, &ucid->mnc, &ucid->lac, &ucid->rac, &ucid->sac, &ucid->cid);
+	int rc;
+	char buf[4];
+	const char *pos = instr;
+	const char *end;
+
+	/* We want to use struct umts_cell_id as hashtable key. If it ever happens to contain any padding bytes, make
+	 * sure everything is deterministically zero. */
+	memset(ucid, 0, sizeof(*ucid));
+
+	/* read MCC */
+	end = strchr(pos, '-');
+	if (!end || end <= pos || (end - pos) >= sizeof(buf))
+		return -EINVAL;
+	osmo_strlcpy(buf, pos, end - pos + 1);
+	if (osmo_mcc_from_str(buf, &ucid->plmn.mcc))
+		return -EINVAL;
+	pos = end + 1;
+
+	/* read MNC -- here the number of leading zeros matters. */
+	end = strchr(pos, '-');
+	if (!end || end == pos || (end - pos) >= sizeof(buf))
+		return -EINVAL;
+	osmo_strlcpy(buf, pos, end - pos + 1);
+	if (osmo_mnc_from_str(buf, &ucid->plmn.mnc, &ucid->plmn.mnc_3_digits))
+		return -EINVAL;
+	pos = end + 1;
+
+	/* parse the rest, where leading zeros do not matter */
+	rc = sscanf(pos, "L%hu-R%hu-S%hu-C%u", &ucid->lac, &ucid->rac, &ucid->sac, &ucid->cid);
 	if (rc < 0)
 		return -errno;
 
-	if (rc != 6)
-		return -EINVAL;
-
-	if (ucid->mcc > 999)
-		return -EINVAL;
-
-	if (ucid->mnc > 999)
+	if (rc != 4)
 		return -EINVAL;
 
 	if (ucid->lac == 0 || ucid->lac == 0xffff)
@@ -291,7 +247,7 @@ const char *hnb_context_name(struct hnb_context *ctx)
 	if (g_hnbgw->config.log_prefix_hnb_id)
 		result = talloc_asprintf(OTC_SELECT, "%s %s", result, ctx->identity_info);
 	else
-		result = talloc_asprintf(OTC_SELECT, "%s %s", result, umts_cell_id_name(&ctx->id));
+		result = talloc_asprintf(OTC_SELECT, "%s %s", result, umts_cell_id_to_str(&ctx->id));
 	return result;
 }
 
@@ -304,19 +260,20 @@ void hnb_context_release_ue_state(struct hnb_context *ctx)
 		context_map_hnb_released(map);
 		/* hnbgw_context_map will remove itself from lists when it is ready. */
 	}
-	ue_context_free_by_hnb(ctx);
 }
 
 void hnb_context_release(struct hnb_context *ctx)
 {
 	struct hnbgw_context_map *map;
-	struct timespec tp;
-	int rc;
 
 	LOGHNB(ctx, DMAIN, LOGL_INFO, "Releasing HNB context\n");
 
-	rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
-	ctx->persistent->updowntime = (rc < 0) ? 0 : tp.tv_sec;
+	if (ctx->persistent) {
+		struct timespec tp;
+		int rc;
+		rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
+		ctx->persistent->updowntime = (rc < 0) ? 0 : tp.tv_sec;
+	}
 
 	/* remove from the list of HNB contexts */
 	llist_del(&ctx->list);
@@ -341,7 +298,7 @@ void hnb_context_release(struct hnb_context *ctx)
 
 	/* remove back reference from hnb_persistent to context */
 	if (ctx->persistent)
-		ctx->persistent->ctx = NULL;
+		hnb_persistent_deregistered(ctx->persistent);
 
 	talloc_free(ctx);
 }
@@ -396,11 +353,14 @@ const struct rate_ctr_desc hnb_ctr_description[] = {
 	[HNB_CTR_RANAP_CS_RAB_MOD_FAIL] = {
 		"ranap:cs:rab_mod:fail", "CS RAB Modifications failed" },
 
-
 	[HNB_CTR_RANAP_PS_RAB_REL_REQ] = {
-		"ranap:ps:rab_rel:req", "PS RAB Release requested" },
+		"ranap:ps:rab_rel:req:normal", "PS RAB Release requested (by CN), normal" },
 	[HNB_CTR_RANAP_CS_RAB_REL_REQ] = {
-		"ranap:cs:rab_rel:req", "CS RAB Release requested" },
+		"ranap:cs:rab_rel:req:normal", "CS RAB Release requested (by CN), normal" },
+	[HNB_CTR_RANAP_PS_RAB_REL_REQ_ABNORMAL] = {
+		"ranap:ps:rab_rel:req:abnormal", "PS RAB Release requested (by CN), abnormal" },
+	[HNB_CTR_RANAP_CS_RAB_REL_REQ_ABNORMAL] = {
+		"ranap:cs:rab_rel:req:abnormal", "CS RAB Release requested (by CN), abnormal" },
 
 	[HNB_CTR_RANAP_PS_RAB_REL_CNF] = {
 		"ranap:ps:rab_rel:cnf", "PS RAB Release confirmed" },
@@ -413,9 +373,13 @@ const struct rate_ctr_desc hnb_ctr_description[] = {
 		"ranap:cs:rab_rel:fail", "CS RAB Release failed" },
 
 	[HNB_CTR_RANAP_PS_RAB_REL_IMPLICIT] = {
-		"ranap:ps:rab_rel:implicit", "PS RAB Release implicit (during Iu Release)" },
+		"ranap:ps:rab_rel:implicit:normal", "PS RAB Release implicit (during Iu Release), normal" },
 	[HNB_CTR_RANAP_CS_RAB_REL_IMPLICIT] = {
-		"ranap:cs:rab_rel:implicit", "CS RAB Release implicit (during Iu Release)" },
+		"ranap:cs:rab_rel:implicit:normal", "CS RAB Release implicit (during Iu Release), normal" },
+	[HNB_CTR_RANAP_PS_RAB_REL_IMPLICIT_ABNORMAL] = {
+		"ranap:ps:rab_rel:implicit:abnormal", "PS RAB Release implicit (during Iu Release), abnormal" },
+	[HNB_CTR_RANAP_CS_RAB_REL_IMPLICIT_ABNORMAL] = {
+		"ranap:cs:rab_rel:implicit:abnormal", "CS RAB Release implicit (during Iu Release), abnormal" },
 
 	[HNB_CTR_RUA_ERR_IND] = {
 		"rua:error_ind", "Received RUA Error Indications" },
@@ -455,6 +419,44 @@ const struct rate_ctr_desc hnb_ctr_description[] = {
 
 	[HNB_CTR_RAB_ACTIVE_MILLISECONDS_TOTAL] = {
 		"rab:cs:active_milliseconds:total", "Cumulative number of milliseconds of CS RAB activity" },
+
+	[HNB_CTR_DTAP_CS_LU_REQ] = { "dtap:cs:location_update:req", "CS Location Update Requests" },
+	[HNB_CTR_DTAP_CS_LU_ACC] = { "dtap:cs:location_update:accept", "CS Location Update Accepts" },
+	[HNB_CTR_DTAP_CS_LU_REJ] = { "dtap:cs:location_update:reject", "CS Location Update Rejects" },
+
+	[HNB_CTR_DTAP_PS_ATT_REQ] = { "dtap:ps:attach:req", "PS Attach Requests" },
+	[HNB_CTR_DTAP_PS_ATT_ACK] = { "dtap:ps:attach:accept", "PS Attach Accepts" },
+	[HNB_CTR_DTAP_PS_ATT_REJ] = { "dtap:ps:attach:reject", "PS Attach Rejects" },
+
+	[HNB_CTR_DTAP_PS_RAU_REQ] = { "dtap:ps:routing_area_update:req", "PS Routing Area Update Requests" },
+	[HNB_CTR_DTAP_PS_RAU_ACK] = { "dtap:ps:routing_area_update:accept", "PS Routing Area Update Accepts" },
+	[HNB_CTR_DTAP_PS_RAU_REJ] = { "dtap:ps:routing_area_update:reject", "PS Routing Area Update Rejects" },
+
+	[HNB_CTR_GTPU_PACKETS_UL] = {
+		"gtpu:packets:ul",
+		"Count of GTP-U packets received from the HNB",
+	},
+	[HNB_CTR_GTPU_TOTAL_BYTES_UL] = {
+		"gtpu:total_bytes:ul",
+		"Count of total GTP-U bytes received from the HNB, including the GTP-U/UDP/IP headers",
+	},
+	[HNB_CTR_GTPU_UE_BYTES_UL] = {
+		"gtpu:ue_bytes:ul",
+		"Assuming an IP header length of 20 bytes, GTP-U bytes received from the HNB, excluding the GTP-U/UDP/IP headers",
+	},
+	[HNB_CTR_GTPU_PACKETS_DL] = {
+		"gtpu:packets:dl",
+		"Count of GTP-U packets sent to the HNB",
+	},
+	[HNB_CTR_GTPU_TOTAL_BYTES_DL] = {
+		"gtpu:total_bytes:dl",
+		"Count of total GTP-U bytes sent to the HNB, including the GTP-U/UDP/IP headers",
+	},
+	[HNB_CTR_GTPU_UE_BYTES_DL] = {
+		"gtpu:ue_bytes:dl",
+		"Assuming an IP header length of 20 bytes, GTP-U bytes sent to the HNB, excluding the GTP-U/UDP/IP headers",
+	},
+
 };
 
 const struct rate_ctr_group_desc hnb_ctrg_desc = {
@@ -477,24 +479,54 @@ const struct osmo_stat_item_group_desc hnb_statg_desc = {
 	.item_desc = hnb_stat_desc,
 };
 
+static void hnb_persistent_disconnected_timeout_cb(void *data)
+{
+	hnb_persistent_free(data);
+}
+
+static void hnb_persistent_disconnected_timeout_schedule(struct hnb_persistent *hnbp)
+{
+	unsigned long period_s = osmo_tdef_get(hnbgw_T_defs, -35, OSMO_TDEF_S, 60*60*24*7);
+	if (period_s < 1) {
+		LOG_HNBP(hnbp, LOGL_INFO,
+			 "timer X35 is zero, not setting a disconnected timeout for this hnb-persistent instance.\n");
+		return;
+	}
+	/* It is fine if the timer is already active, osmo_timer_del() is done implicitly by the osmo_timer API. */
+	osmo_timer_setup(&hnbp->disconnected_timeout, hnb_persistent_disconnected_timeout_cb, hnbp);
+	osmo_timer_schedule(&hnbp->disconnected_timeout, period_s, 0);
+}
+
+static unsigned int g_hnbp_ctr_id;
+
 struct hnb_persistent *hnb_persistent_alloc(const struct umts_cell_id *id)
 {
 	struct hnb_persistent *hnbp = talloc_zero(g_hnbgw, struct hnb_persistent);
+	unsigned int ctr_id = g_hnbp_ctr_id++;
 	if (!hnbp)
 		return NULL;
 
 	hnbp->id = *id;
-	hnbp->id_str = talloc_strdup(hnbp, umts_cell_id_name(id));
-	hnbp->ctrs = rate_ctr_group_alloc(hnbp, &hnb_ctrg_desc, 0);
+	hnbp->id_str = talloc_strdup(hnbp, umts_cell_id_to_str(id));
+	hnbp->ctrs = rate_ctr_group_alloc(hnbp, &hnb_ctrg_desc, ctr_id);
 	if (!hnbp->ctrs)
 		goto out_free;
 	rate_ctr_group_set_name(hnbp->ctrs, hnbp->id_str);
-	hnbp->statg = osmo_stat_item_group_alloc(hnbp, &hnb_statg_desc, 0);
+	hnbp->statg = osmo_stat_item_group_alloc(hnbp, &hnb_statg_desc, ctr_id);
 	if (!hnbp->statg)
 		goto out_free_ctrs;
 	osmo_stat_item_group_set_name(hnbp->statg, hnbp->id_str);
 
 	llist_add(&hnbp->list, &g_hnbgw->hnb_persistent_list);
+	hash_add(g_hnbgw->hnb_persistent_by_id, &hnbp->node_by_id, umts_cell_id_hash(&hnbp->id));
+
+	if (g_hnbgw->nft_kpi.active)
+		nft_kpi_hnb_persistent_add(hnbp);
+
+	/* Normally the disconnected timer runs only when the hNodeB is not currently connected on Iuh. This here is paranoia:
+	 * In case we have to HNBAP HNB Register Reject, the disconnected timer should be active on this unused hnbp.
+	 * On success, hnb_persistent_registered() will stop the disconnected timer directly after this. */
+	hnb_persistent_disconnected_timeout_schedule(hnbp);
 
 	return hnbp;
 
@@ -508,19 +540,93 @@ out_free:
 struct hnb_persistent *hnb_persistent_find_by_id(const struct umts_cell_id *id)
 {
 	struct hnb_persistent *hnbp;
-
-	llist_for_each_entry(hnbp, &g_hnbgw->hnb_persistent_list, list) {
+	uint32_t id_hash = umts_cell_id_hash(id);
+	hash_for_each_possible (g_hnbgw->hnb_persistent_by_id, hnbp, node_by_id, id_hash) {
 		if (umts_cell_id_equal(&hnbp->id, id))
 			return hnbp;
 	}
-
 	return NULL;
 }
 
+/* Read the peer's remote IP address from the Iuh conn's fd, and set up GTP-U counters for that remote address. */
+static void hnb_persistent_update_remote_addr(struct hnb_persistent *hnbp)
+{
+	socklen_t socklen;
+	struct osmo_sockaddr osa;
+	struct osmo_sockaddr_str remote_str;
+	int fd;
+
+	fd = osmo_stream_srv_get_fd(hnbp->ctx->conn);
+	if (fd < 0) {
+		LOG_HNBP(hnbp, LOGL_ERROR, "no active socket fd, cannot set up traffic counters\n");
+		return;
+	}
+
+	socklen = sizeof(struct osmo_sockaddr);
+	if (getpeername(fd, &osa.u.sa, &socklen)) {
+		LOG_HNBP(hnbp, LOGL_ERROR, "cannot read remote address, cannot set up traffic counters\n");
+		return;
+	}
+	if (osmo_sockaddr_str_from_osa(&remote_str, &osa)) {
+		LOG_HNBP(hnbp, LOGL_ERROR, "cannot parse remote address, cannot set up traffic counters\n");
+		return;
+	}
+
+	/* We got the remote address from the Iuh link (RUA), and now we are blatantly assuming that the hNodeB has its
+	 * GTP endpoint on the same IP address, just with UDP port 2152 (the fixed GTP port as per 3GPP spec). */
+	remote_str.port = 2152;
+
+	if (nft_kpi_hnb_start(hnbp, &remote_str))
+		LOG_HNBP(hnbp, LOGL_ERROR, "failed to set up traffic counters\n");
+}
+
+/* Whenever HNBAP registers a HNB, hnbgw_hnbap.c calls this function to let the hnb_persistent update its state to the
+ * (new) remote address being active. When calling this function, a hnbp->ctx should be present, with an active
+ * osmo_stream_srv conn. */
+void hnb_persistent_registered(struct hnb_persistent *hnbp)
+{
+	if (!hnbp->ctx) {
+		LOG_HNBP(hnbp, LOGL_ERROR, "hnb_persistent_registered() invoked, but there is no hnb_ctx\n");
+		return;
+	}
+
+	/* The hNodeB is now connected, i.e. not disconnected. */
+	osmo_timer_del(&hnbp->disconnected_timeout);
+
+	/* start counting traffic */
+	if (g_hnbgw->nft_kpi.active)
+		hnb_persistent_update_remote_addr(hnbp);
+}
+
+/* Whenever a HNB is regarded as no longer registered (HNBAP HNB De-Register, or the Iuh link drops), this function is
+ * called to to let the hnb_persistent update its state to the hNodeB being disconnected. Clear the ctx->persistent and
+ * hnbp->ctx relations; do not delete the hnb_persistent instance. */
+void hnb_persistent_deregistered(struct hnb_persistent *hnbp)
+{
+	/* clear out cross references of hnb_context and hnb_persistent */
+	if (hnbp->ctx) {
+		if (hnbp->ctx->persistent == hnbp)
+			hnbp->ctx->persistent = NULL;
+		hnbp->ctx = NULL;
+	}
+
+	/* stop counting traffic */
+	nft_kpi_hnb_stop(hnbp);
+
+	/* The hNodeB is now disconnected. Clear out hnb_persistent when the disconnected timeout has passed. */
+	hnb_persistent_disconnected_timeout_schedule(hnbp);
+}
+
 void hnb_persistent_free(struct hnb_persistent *hnbp)
 {
 	/* FIXME: check if in use? */
+	osmo_timer_del(&hnbp->disconnected_timeout);
+	nft_kpi_hnb_stop(hnbp);
+	nft_kpi_hnb_persistent_remove(hnbp);
+	osmo_stat_item_group_free(hnbp->statg);
+	rate_ctr_group_free(hnbp->ctrs);
 	llist_del(&hnbp->list);
+	hash_del(&hnbp->node_by_id);
 	talloc_free(hnbp);
 }
 
@@ -845,6 +951,11 @@ static const struct log_info_cat hnbgw_log_cat[] = {
 		.color = OSMO_LOGCOLOR_DARKYELLOW,
 		.description = "Core Network side (via SCCP)",
 	},
+	[DNFT] = {
+		.name = "DNFT", .loglevel = LOGL_NOTICE, .enabled = 1,
+		.color = OSMO_LOGCOLOR_BLUE,
+		.description = "nftables interaction for retrieving stats",
+	},
 };
 
 const struct log_info hnbgw_log_info = {
@@ -875,8 +986,10 @@ void g_hnbgw_alloc(void *ctx)
 
 	g_hnbgw->next_ue_ctx_id = 23;
 	INIT_LLIST_HEAD(&g_hnbgw->hnb_list);
+
 	INIT_LLIST_HEAD(&g_hnbgw->hnb_persistent_list);
-	INIT_LLIST_HEAD(&g_hnbgw->ue_list);
+	hash_init(g_hnbgw->hnb_persistent_by_id);
+
 	INIT_LLIST_HEAD(&g_hnbgw->sccp.users);
 
 	g_hnbgw->mgw_pool = mgcp_client_pool_alloc(g_hnbgw);
